import warnings
import ConfigSpace as CS
import numpy as np
from scipy import stats

from xbbo.configspace.abstract_feature_space import (
    AbstractFeatureSpace,
    Identity,
    Ord2Uniform,
    Cat2Onehot,
    U2gaussian,
    Category,
    Ordinal,
    U2Onehot
)

class FeatureSpace_discrete_all_oneHot(AbstractFeatureSpace):
    '''
    sparse array <=> feature space
    all variable i.i.d ~ U(0, 1)
    '''

    def __init__(self, discrete_degree):
        super().__init__()
        # self.dtypes_idx_map = dtypes_idx_map
        # self.discrete_degree = discrete_degree
        self.discrete_degree = discrete_degree
        self.unif = U2Onehot()
        self.cat = Cat2Onehot()
        self.ord = Cat2Onehot()
        self.features_ = []
        self.src_ids = []
        self.trg_ids = []
        self.dtypes_idx_map = {
            'cat': Array_idx_map(),
            'int': Array_idx_map(),
            'float': Array_idx_map(),
            'ord': Array_idx_map(),
        }

        nums, ords, cats, size_sparse, size_dense, categories = self._get_mappings()
        if nums:
            self.dtypes_idx_map['float'].src_ids, self.dtypes_idx_map['float'].trg_ids, self.dtypes_idx_map[
                'float'].cat_sizes = \
                map(np.uintp, zip(*nums))
            self.dtypes_idx_map['float'].cats = nums
        if ords:
            self.dtypes_idx_map['ord'].src_ids, self.dtypes_idx_map['ord'].trg_ids, self.dtypes_idx_map[
                'ord'].cat_sizes = \
                map(np.uintp, zip(*ords))
            self.dtypes_idx_map['ord'].cats = ords

        if cats:
            self.dtypes_idx_map['cat'].src_ids, self.dtypes_idx_map['cat'].trg_ids, self.dtypes_idx_map[
                'cat'].cat_sizes = \
                map(np.uintp, zip(*cats))
            self.dtypes_idx_map['cat'].cats = cats

        # self.nums = nums
        # self.cats = cats
        self.categories = np.asarray(categories)
        self.sparse_dimension = size_sparse
        self.dense_dimension = size_dense

    def _get_mappings(self):

        nums = []
        # nums_float = []
        # nums_ord = []
        cats = []
        ords = []
        categories = []
        src_ind = trg_ind = 0
        for src_ind, hp in enumerate(self.space.get_hyperparameters()):
            if isinstance(hp, CS.CategoricalHyperparameter):
                cat_size = hp.num_choices
                cats.append((src_ind, trg_ind, cat_size))
                trg_ind += cat_size
                categories.append(len(hp.choices))
            elif isinstance(hp, (CS.UniformIntegerHyperparameter,CS.UniformFloatHyperparameter)):
                nums.append((src_ind, trg_ind, self.discrete_degree))
                categories.append(self.discrete_degree)
                trg_ind += self.discrete_degree
            elif isinstance(hp, CS.OrdinalHyperparameter):
                categories.append(len(hp.sequence))
                cat_size = hp.num_elements
                ords.append((src_ind, trg_ind, cat_size))
                trg_ind += cat_size
            # elif isinstance(hp, (CS.UniformIntegerHyperparameter,
            #                      CS.UniformFloatHyperparameter,
            #                      CS.OrdinalHyperparameter)):
            #     nums.append((src_ind, trg_ind))

            else:
                raise NotImplementedError(
                    "Only hyperparameters of types "
                    "`CategoricalHyperparameter`, "
                    "`UniformIntegerHyperparameter`, "
                    "`OrdinalHyperparameter`, "
                    "`UniformFloatHyperparameter` are supported!")

        size_sparse = src_ind + 1
        size_dense = trg_ind


        return nums, ords, cats, size_sparse, size_dense, categories


    def feature_to_array(self, x_feature, sparse_dim):
        '''
        return sparse array for construct Configurations
        '''
        assert not (self.dtypes_idx_map is None)
        x_array = np.empty(shape=(sparse_dim))

        for dtype, array_idx_map in self.dtypes_idx_map.items():
            if array_idx_map.src_ids is None:
                continue
            if dtype == 'cat':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.cat.feature_to_sparse_array(
                        x_feature[trg_ind:trg_ind + size], size
                    )
            elif dtype == 'ord':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.ord.feature_to_sparse_array(
                        x_feature[trg_ind:trg_ind + size], size
                    )
                # x_array[array_idx_map.src_ids] = \
                #     self.ord.feature_to_sparse_array(
                #         x_feature[array_idx_map.trg_ids],
                #     )
            elif dtype in ('float', 'int'):
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.unif.feature_to_sparse_array(
                        x_feature[trg_ind:trg_ind + size], size
                    )
            else:
                pass
        return x_array

    def array_to_feature(self, array, dense_dim): # TODO
        warnings.warn("This method may no reason be called?")
        assert not (self.dtypes_idx_map is None)
        feature = np.zeros(shape=(dense_dim))
        for dtype, array_idx_map in self.dtypes_idx_map.items():
            # if array_idx_map.src_ids:
            if array_idx_map.src_ids is None:
                continue
            if dtype == 'cat':
                # for src_ind, trg_ind, size in (array_idx_map.cats):
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind:trg_ind+size] = \
                        self.cat.sparse_array_to_feature(
                            array[src_ind], size
                        )
            elif dtype == 'ord':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind:trg_ind+size] = \
                        self.ord.sparse_array_to_feature(
                            array[src_ind], size
                        )
            elif dtype in ('float', 'int'):
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind:trg_ind+size] = \
                        self.unif.sparse_array_to_feature(
                            array[src_ind], size
                        )
            else:
                pass
        return feature

    def record_feature(self, feature):
        self.features_.append(feature)

class FeatureSpace_uniform(AbstractFeatureSpace):
    '''
    sparse array <=> feature space
    all variable i.i.d ~ U(0, 1)
    '''

    def __init__(self, dtypes_idx_map):
        super().__init__()
        self.dtypes_idx_map = dtypes_idx_map
        self.unif = Identity()
        self.cat = Cat2Onehot()
        self.ord = Ord2Uniform()
        self.features_ = []


    def feature_to_array(self, x_feature, sparse_dim):
        '''
        return sparse array for construct Configurations
        '''
        assert not (self.dtypes_idx_map is None)
        x_array = np.empty(shape=(sparse_dim))
        for dtype, array_idx_map in self.dtypes_idx_map.items():
            if array_idx_map.src_ids is None:
                continue
            if dtype == 'cat':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.cat.feature_to_sparse_array(
                        x_feature[trg_ind:trg_ind + size], size
                    )
            elif dtype == 'ord':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.ord.feature_to_sparse_array(
                        x_feature[trg_ind], size
                    )
                # x_array[array_idx_map.src_ids] = \
                #     self.ord.feature_to_sparse_array(
                #         x_feature[array_idx_map.trg_ids],
                #     )
            elif dtype in ('float', 'int'):
                x_array[array_idx_map.src_ids] = \
                    self.unif.feature_to_sparse_array(
                        x_feature[array_idx_map.trg_ids]
                    )
            else:
                pass
        return x_array

    def array_to_feature(self, array, dense_dim): # TODO
        warnings.warn("This method may no reason be called?")
        assert not (self.dtypes_idx_map is None)
        feature = np.zeros(shape=(dense_dim))
        for dtype, array_idx_map in self.dtypes_idx_map.items():
            # if array_idx_map.src_ids:
            if array_idx_map.src_ids is None:
                continue
            if dtype == 'cat':
                # for src_ind, trg_ind, size in (array_idx_map.cats):
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind:trg_ind+size] = \
                        self.cat.sparse_array_to_feature(
                            array[src_ind], size
                        )
            elif dtype == 'ord':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind] = \
                        self.ord.sparse_array_to_feature(
                            array[src_ind], size
                        )
            elif dtype in ('float', 'int'):
                feature[array_idx_map.trg_ids] = \
                    self.unif.sparse_array_to_feature(
                        array[array_idx_map.src_ids]
                    )
            else:
                pass
        return feature

    def record_feature(self, feature):
        self.features_.append(feature)

class FeatureSpace_gaussian(AbstractFeatureSpace):

    '''
    sparse array <=> feature space
    all variable i.i.d ~ Normal distribution N(0, 1)
    '''

    def __init__(self, dtypes_idx_map):
        super().__init__()
        self.dtypes_idx_map = dtypes_idx_map
        self.unif = U2gaussian()
        self.cat = Category()
        self.ord = Ordinal()
        self.features_ = []


    def feature_to_array(self, x_feature, sparse_dim):
        '''
        return sparse array for construct Configurations
        '''
        assert not (self.dtypes_idx_map is None)
        x_array = np.empty(shape=(sparse_dim))
        for dtype, array_idx_map in self.dtypes_idx_map.items():
            if array_idx_map.src_ids is None:
                continue
            if dtype == 'cat':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.cat.feature_to_sparse_array(
                        x_feature[trg_ind:trg_ind + size], size
                    )
            elif dtype == 'ord':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    x_array[src_ind] = \
                        self.ord.feature_to_sparse_array(
                        x_feature[trg_ind], size
                    )
                # x_array[array_idx_map.src_ids] = \
                #     self.ord.feature_to_sparse_array(
                #         x_feature[array_idx_map.trg_ids],
                #     )
            elif dtype in ('float', 'int'):
                x_array[array_idx_map.src_ids] = \
                    self.unif.feature_to_sparse_array(
                        x_feature[array_idx_map.trg_ids]
                    )
            else:
                pass
        return x_array

    def array_to_feature(self, array, dense_dim): # TODO
        warnings.warn("This method may no reason be called?")
        assert not (self.dtypes_idx_map is None)
        feature = np.zeros(shape=(dense_dim))
        for dtype, array_idx_map in self.dtypes_idx_map.items():
            # if array_idx_map.src_ids:
            if array_idx_map.src_ids is None:
                continue
            if dtype == 'cat':
                # for src_ind, trg_ind, size in (array_idx_map.cats):
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind:trg_ind+size] = \
                        self.cat.sparse_array_to_feature(
                            array[src_ind], size
                        )
            elif dtype == 'ord':
                for src_ind, trg_ind, size in (array_idx_map.cats):
                    feature[trg_ind] = \
                        self.ord.sparse_array_to_feature(
                            array[src_ind], size
                        )
            elif dtype in ('float', 'int'):
                feature[array_idx_map.trg_ids] = \
                    self.unif.sparse_array_to_feature(
                        array[array_idx_map.src_ids]
                    )
            else:
                pass
        return feature

    def record_feature(self, feature):
        self.features_.append(feature)

class Uniform2Gaussian(AbstractFeatureSpace):
    def __init__(self):
        super().__init__()
    
    def array_to_feature(self, array):
        return stats.norm.ppf(array)

    def feature_to_array(self, feature):
        return stats.norm.cdf(feature)


from typing import Optional

import numpy as np
from scipy import stats

from xbbo.acquisition_function.base import AbstractAcquisitionFunction

class TAF_AcqFunc(AbstractAcquisitionFunction):
    def __init__(self, surrogate_model, base_models, rng, rho=None):
        self.par = 0.0
        self.base_models = base_models
        self.rng = rng
        self.rho = rho
        super().__init__(surrogate_model)

    def argmax(self, candidates):
        print('depreciate!')
        # best_ei = -1
        # # best_candidate = []
        # candidates_rm_id = []
        # y_hats = list(zip(*surrogate.predict_with_sigma(candidates)))
        scores = self.__call__(candidates)

        return candidates[self.rng.choice(np.where(scores == scores.max())[0])]

    def update_weight(self, w, rho=None):
        assert w.min() >= 0
        if rho:
            self.w = np.array(w, dtype='float')
            self.selfWeight = rho
        else:
            self.selfWeight = w[-1]
            self.w = np.array(w[:-1],dtype='float')
        #non_zero_weight_indices = (self.w**2 > 0).nonzero()[0]
        non_zero_weight_indices = self.w.nonzero()[0]
        self.pre_weight_model = []
        self.base_incuments = []
        for idx in non_zero_weight_indices:
            self.pre_weight_model.append(self.base_models[idx])
            self.base_incuments.append(self._base_incuments[idx])
        self.weight = self.w[non_zero_weight_indices]
        norm_coeff = (self.weight.sum() + self.selfWeight)
        self.weight /= norm_coeff
        self.selfWeight /= norm_coeff
        
        

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Computes the EI value and its derivatives.

        Parameters
        ----------
        X: np.ndarray(N, D), The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        Returns
        -------
        np.ndarray(N,1)
            Expected Improvement of X
        """
        if len(X.shape) == 1:
            X = X[:, np.newaetas]

        m, v = self.surrogate_model.predict_marginalized_over_instances(X)
        s = np.sqrt(v)

        if self.y_best is None:
            raise ValueError('No current best specified. Call update('
                             'eta=<int>) to inform the acquisition function '
                             'about the current best value.')

        def calculate_f():
            z = (self.y_best - m - self.par) / s
            return (self.y_best - m -
                    self.par) * stats.norm.cdf(z) + s * stats.norm.pdf(z)

        if np.any(s == 0.0):
            # if std is zero, we have observed x on all instances
            # using a RF, std should be never exactly 0.0
            # Avoid zero division by setting all zeros in s to one.
            # Consider the corresponding results in f to be zero.
            s_copy = np.copy(s)
            s[s_copy == 0.0] = 1.0
            f = calculate_f()
            f[s_copy == 0.0] = 0.0
        else:
            f = calculate_f()
        if (f < 0).any():
            raise ValueError(
                "Expected Improvement is smaller than 0 for at least one "
                "sample.")
        # m, v = self.surrogate_model.predict_marginalized_over_instances(X)
        # denominator = self.selfWeight
        f *= self.selfWeight
        for d in range(len(self.weight)):
            f += self.weight[d] * (self.base_incuments[d] - self.pre_weight_model[d].predict(
                X, None)[0]).clip(0)
            # denominator += self.weight[d]

        return f



class TAF_with_RGPE_rank():
    def __init__(
        self,
        gps,
    ):
        self.gps = gps
        self.xi = 0.1

    def _getEI(self, mu, sigma, y_best):  # minimize
        z = (-y_best + mu - self.xi) / sigma
        ei = (-y_best + mu -
              self.xi) * stats.norm.cdf(z) + sigma * stats.norm.pdf(z)
        return ei

    def argmax(self,
               y_best,
               surrogate,
               candidates,
               similarity,
               old_ybests,
               rho=0.75):
        best_ei = -1
        best_candidate = []
        candidates_rm_id = []
        for i, candidate in enumerate(candidates):
            # if not hasattr(surrogate, 'X'):
            if not surrogate.is_fited:
                y_hat = 0, 1000
            else:
                y_hat = surrogate.predict_with_sigma(candidate)
            denominator = rho
            ei = self._getEI(y_hat[0], y_hat[1], y_best).item() * rho
            for d in range(len(similarity)):
                # mu, sigma = self.gps[d].cached_predict_with_sigma(candidate)
                # ei += similarity[d] * max(np.random.normal(mu, sigma)-self.old_Ybests[d], 0)
                mu = self.gps[d].cached_predict(candidate)  # TODO
                ei += similarity[d] * max(mu - old_ybests[d], 0)
                # ei += similarity[d] * max(mu-y_best, 0)
                # mu = self.gps[d].cached_predict(candidate)
                # ei += similarity[d] * max(mu-self.old_Ybests[d][len_h], 0)
                denominator += similarity[d]
            ei /= denominator
            if ei > best_ei:
                best_ei = ei
                best_candidate = [candidate]
                candidates_rm_id = [i]
            elif ei == best_ei:
                best_candidate.append(candidate)
                candidates_rm_id.append(i)

        assert best_candidate
        idx = np.random.choice(len(best_candidate))
        # TODO: this
        # for d in range(len(old_ybests)):
        #     old_ybests[d] = min(old_ybests[d], self.gps[d].cached_predict(best_candidate[idx]))
        return (best_candidate)[idx], candidates_rm_id[idx]


import math
import warnings

import numpy as np
import scipy.stats as sps

VERY_SMALL_NUMBER = 1e-10

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class Prior(object):

    def __init__(self, rng: np.random.RandomState):
        """
        Abstract base class to define the interface for priors
        of GP hyperparameter.

        This class is adapted from RoBO:

        Klein, A. and Falkner, S. and Mansur, N. and Hutter, F.
        RoBO: A Flexible and Robust Bayesian Optimization Framework in Python
        In: NIPS 2017 Bayesian Optimization Workshop

        [16.04.2019]: Whenever lnprob or the gradient is computed for a scalar input, we use math.* rather than np.*

        Parameters
        ----------
        rng: np.random.RandomState
            Random number generator

        """
        if rng is None:
            raise ValueError('Argument rng must not be `None`.')
        self.rng = rng

    def lnprob(self, theta: float) -> float:
        """
        Return the log probability of theta.

        Theta must be on a log scale! This method exponentiates theta and calls ``self._lnprob``.

        Parameters
        ----------
        theta : float
            Hyperparameter configuration in log space.

        Returns
        -------
        float
            The log probability of theta
        """
        return self._lnprob(np.exp(theta))

    def _lnprob(self, theta: float) -> float:
        """
        Return the log probability of theta.

        Theta must be on the original scale.

        Parameters
        ----------
        theta : float
            Hyperparameter configuration on the original scale.

        Returns
        -------
        float
            The log probability of theta
        """
        raise NotImplementedError()

    def sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Returns ``n_samples`` from the prior.

        All samples are on a log scale. This method calls ``self._sample_from_prior`` and applies a log transformation
        to the obtained values.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """

        if np.ndim(n_samples) != 0:
            raise ValueError('argument n_samples needs to be a scalar (is %s)' % n_samples)
        if n_samples <= 0:
            raise ValueError('argument n_samples needs to be positive (is %d)' % n_samples)

        sample = np.log(self._sample_from_prior(n_samples=n_samples))

        if np.any(~np.isfinite(sample)):
            raise ValueError('Sample %s from prior %s contains infinite values!' % (sample, self))

        return sample

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Returns ``n_samples`` from the prior.

        All samples are on a original scale.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """
        raise NotImplementedError()

    def gradient(self, theta: float) -> float:
        """
        Computes the gradient of the prior with respect to theta.

        Theta must be on the original scale.

        Parameters
        ----------
        theta : float
            Hyperparameter configuration in log space

        Returns
        -------
        float
            The gradient of the prior at theta.
        """
        return self._gradient(np.exp(theta))

    def _gradient(self, theta: float) -> float:
        """
        Computes the gradient of the prior with respect to theta.

        Parameters
        ----------
        theta : float
            Hyperparameter configuration in the original space space

        Returns
        -------
        float
            The gradient of the prior at theta.
        """
        raise NotImplementedError()


class TophatPrior(Prior):

    def __init__(self, lower_bound: float, upper_bound: float, rng: np.random.RandomState):
        """
        Tophat prior as it used in the original spearmint code.

        This class is adapted from RoBO:

        Klein, A. and Falkner, S. and Mansur, N. and Hutter, F.
        RoBO: A Flexible and Robust Bayesian Optimization Framework in Python
        In: NIPS 2017 Bayesian Optimization Workshop

        Parameters
        ----------
        lower_bound : float
            Lower bound of the prior. In original scale.
        upper_bound : float
            Upper bound of the prior. In original scale.
        rng: np.random.RandomState
            Random number generator
        """

        super().__init__(rng)
        self.min = lower_bound
        self._log_min = np.log(lower_bound)
        self.max = upper_bound
        self._log_max = np.log(upper_bound)
        if not (self.max > self.min):
            raise Exception("Upper bound of Tophat prior must be greater than the lower bound!")

    def _lnprob(self, theta: float) -> float:
        """
        Return the log probability of theta.

        Parameters
        ----------
        theta : float
            A hyperparameter configuration

        Returns
        -------
        float
        """
        if theta < self.min or theta > self.max:
            return -np.inf
        else:
            return 0

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Return ``n_samples`` from the prior.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """

        if np.ndim(n_samples) != 0:
            raise ValueError('argument n_samples needs to be a scalar (is %s)' % n_samples)
        if n_samples <= 0:
            raise ValueError('argument n_samples needs to be positive (is %d)' % n_samples)

        p0 = np.exp(self.rng.uniform(low=self._log_min, high=self._log_max, size=(n_samples, )))
        return p0

    def gradient(self, theta: float) -> float:
        """
        Computes the gradient of the prior with respect to theta.

        Parameters
        ----------
        theta : float
            Hyperparameter configuration in log space

        Returns
        -------
        (D) np.array

            The gradient of the prior at theta.
        """
        return 0


class HorseshoePrior(Prior):

    def __init__(self, scale: float, rng: np.random.RandomState):
        """
        Horseshoe Prior as it is used in spearmint

        This class is adapted from RoBO:

        Klein, A. and Falkner, S. and Mansur, N. and Hutter, F.
        RoBO: A Flexible and Robust Bayesian Optimization Framework in Python
        In: NIPS 2017 Bayesian Optimization Workshop

        Parameters
        ----------
        scale: float
            Scaling parameter. See below how it is influencing the distribution.
        rng: np.random.RandomState
            Random number generator
        """
        super().__init__(rng)
        self.scale = scale
        self.scale_square = scale ** 2

    def _lnprob(self, theta: float) -> float:
        """
        Return the log probability of theta.

        Parameters
        ----------
        theta : (D,) numpy array
            A hyperparameter configuration

        Returns
        -------
        float
        """
        # We computed it exactly as in the original spearmint code, they basically say that there's no analytical form
        # of the horseshoe prior, but that the multiplier is bounded between 2 and 4 and that they used the middle
        # See "The horseshoe estimator for sparse signals" by Carvalho, Poloson and Scott (2010), Equation 1.
        # https://www.jstor.org/stable/25734098
        # Compared to the paper by Carvalho, there's a constant multiplicator missing
        # Compared to Spearmint we first have to undo the log space transformation of the theta
        # Note: "undo log space transformation" is done in parent class
        if theta == 0:
            return np.inf  # POSITIVE infinity (this is the "spike")
        else:
            a = math.log(1 + 3.0 * (self.scale_square / theta**2))
            return math.log(a + VERY_SMALL_NUMBER)

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Returns N samples from the prior.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """

        # This is copied from RoBO - scale is most likely the tau parameter
        lamda = np.abs(self.rng.standard_cauchy(size=n_samples))
        p0 = np.abs(self.rng.randn() * lamda * self.scale)
        return p0

    def _gradient(self, theta: float) -> float:
        """
        Computes the gradient of the prior with
        respect to theta.

        Parameters
        ----------
        theta : (D,) numpy array
            Hyperparameter configuration

        Returns
        -------
        (D) np.array
            The gradient of the prior at theta.
        """
        if theta == 0:
            return np.inf  # POSITIVE infinity (this is the "spike")
        else:
            a = -(6 * self.scale_square)
            b = 3 * self.scale_square + theta**2
            b *= math.log(3 * self.scale_square * theta ** (-2) + 1)
            b = max(b, 1e-14)
            return a / b


class LognormalPrior(Prior):

    def __init__(self, sigma: float, rng: np.random.RandomState, mean: float = 0, ):
        """
        Log normal prior

        This class is adapted from RoBO:

        Klein, A. and Falkner, S. and Mansur, N. and Hutter, F.
        RoBO: A Flexible and Robust Bayesian Optimization Framework in Python
        In: NIPS 2017 Bayesian Optimization Workshop

        Parameters
        ----------
        sigma: float
            Specifies the standard deviation of the normal
            distribution.
        rng: np.random.RandomState
            Random number generator
        mean: float
            Specifies the mean of the normal distribution
        """
        super().__init__(rng)

        if mean != 0:
            raise NotImplementedError(mean)

        self.sigma = sigma
        self.sigma_square = sigma ** 2
        self.mean = mean
        self.sqrt_2_pi = np.sqrt(2 * np.pi)

    def _lnprob(self, theta: float) -> float:
        """
        Return the log probability of theta

        Parameters
        ----------
        theta : float
            A hyperparameter configuration

        Returns
        -------
        float
        """
        if theta <= self.mean:
            return -1e25
        else:
            rval = (
                -(math.log(
                    theta) - self.mean) ** 2 / (2 * self.sigma_square) - math.log(self.sqrt_2_pi * self.sigma * theta)
            )
            return rval

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Returns N samples from the prior.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """

        return self.rng.lognormal(mean=self.mean, sigma=self.sigma, size=n_samples)

    def _gradient(self, theta: float) -> float:
        """
        Computes the gradient of the prior with
        respect to theta.

        Parameters
        ----------
        theta : (D,) numpy array
            Hyperparameter configuration in log space

        Returns
        -------
        (D) np.array
            The gradient of the prior at theta.
        """
        if theta <= 0:
            return 0
        else:
            # derivative of log(1 / (x * s^2 * sqrt(2 pi)) * exp( - 0.5 * (log(x ) / s^2))^2))
            # This is without the mean!!!
            return -(self.sigma_square + math.log(theta)) / (self.sigma_square * (theta)) * theta


class SoftTopHatPrior(Prior):
    def __init__(self, lower_bound: float, upper_bound: float, exponent: float, rng: np.random.RandomState) -> None:
        super().__init__(rng)

        with warnings.catch_warnings():
            warnings.simplefilter('error')
            self.lower_bound = lower_bound
            try:
                self._log_lower_bound = np.log(lower_bound)
            except RuntimeWarning as w:
                if 'invalid value encountered in log' in w.args[0]:
                    raise ValueError('Invalid lower bound %f (cannot compute log)' % lower_bound)
                else:
                    raise w
            self.upper_bound = upper_bound
            try:
                self._log_upper_bound = np.log(upper_bound)
            except RuntimeWarning as w:
                if 'invalid value encountered in log' in w.args[0]:
                    raise ValueError('Invalid lower bound %f (cannot compute log)' % lower_bound)
                else:
                    raise w

        if exponent <= 0:
            raise ValueError('Exponent cannot be less or equal than zero (but is %f)' % exponent)
        self.exponent = exponent

    def lnprob(self, theta: float) -> float:
        # We need to use lnprob here instead of _lnprob to have the squared function work in the logarithmic space,
        # too.
        if np.ndim(theta) == 0:
            if theta < self._log_lower_bound:
                return - ((theta - self._log_lower_bound) ** self.exponent)
            elif theta > self._log_upper_bound:
                return - (self._log_upper_bound - theta) ** self.exponent
            else:
                return 0
        else:
            raise NotImplementedError()

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Returns N samples from the prior.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """

        return np.exp(self.rng.uniform(self._log_lower_bound, self._log_upper_bound, size=(n_samples, )))

    def gradient(self, theta: float) -> float:
        if np.ndim(theta) == 0:
            if theta < self._log_lower_bound:
                return - self.exponent * (theta - self._log_lower_bound)
            elif theta > self._log_upper_bound:
                return self.exponent * (self._log_upper_bound - theta)
            else:
                return 0
        else:
            raise NotImplementedError()

    def __repr__(self) -> str:
        return 'SoftTopHatPrior(lower_bound=%f, upper_bound=%f)' % (self.lower_bound, self.upper_bound)


class GammaPrior(Prior):

    def __init__(self, a: float, scale: float, loc: float, rng: np.random.RandomState):
        """
        Gamma prior

        f(x) = (x-loc)**(a-1) * e**(-(x-loc)) * (1/scale)**a / gamma(a)

        Parameters
        ----------
        a: float > 0
            shape parameter
        scale: float > 0
            scale parameter (1/scale corresponds to parameter p in canonical form)
        loc: float
            mean parameter for the distribution
        rng: np.random.RandomState
            Random number generator
        """
        super().__init__(rng)

        self.a = a
        self.loc = loc
        self.scale = scale

    def _lnprob(self, theta: float) -> float:
        """
        Returns the logpdf of theta.

        Parameters
        ----------
        theta : float
            Hyperparameter configuration

        Returns
        -------
        float
        """
        if np.ndim(theta) != 0:
            raise NotImplementedError()
        return sps.gamma.logpdf(theta, a=self.a, scale=self.scale, loc=self.loc)

    def _sample_from_prior(self, n_samples: int) -> np.ndarray:
        """
        Returns N samples from the prior.

        Parameters
        ----------
        n_samples : int
            The number of samples that will be drawn.

        Returns
        -------
        np.ndarray
        """

        return self.rng.gamma(shape=self.a, scale=self.scale, size=n_samples)

    def _gradient(self, theta: float) -> float:
        """
        As computed by Wolfram Alpha

        Parameters
        ----------
        theta: float
            A hyperparameter configuration

        Returns
        -------
        float
        """
        if np.ndim(theta) == 0:
            # Multiply by theta because of the chain rule...
            return ((self.a - 1) / theta - (1 / self.scale)) * theta
        else:
            raise NotImplementedError()



from abc import ABC, abstractmethod
import time

import numpy as np
import ConfigSpace as CS
from xbbo.configspace.space import DenseConfigurationSpace
from xbbo.core.trials import Trials
from xbbo.utils.constants import Key
# from xbbo.configspace.space import Configurations


class AbstractOptimizer(ABC):
    """Abstract base class for the optimizers in the benchmark. This creates a common API across all packages.
    """

    # Every implementation package needs to specify this static variable, e.g., "primary_import=opentuner"
    primary_import = None

    def __init__(self,
                 space: CS.ConfigurationSpace,
                 encoding_cat='round',
                 encoding_ord='round',
                 seed=42,
                 suggest_limit: float = np.inf,
                 total_time_limit: float = np.inf,
                 cost_limit: float = np.inf,
                 learner_time_limit: float = np.inf,
                 budget_limit: float= np.inf,
                 objective_function=None,
                 **kwargs):
        """Build wrapper class to use an optimizer in benchmark.

        Parameters
        ----------
        api_config : dict-like of dict-like
            Configuration of the optimization variables. See API description.
        """
        assert isinstance(space, CS.ConfigurationSpace)
        if not isinstance(space, DenseConfigurationSpace):
            self.space = DenseConfigurationSpace(space,
                                             encoding_cat=encoding_cat,
                                             encoding_ord=encoding_ord)
        else:
            self.space = space
        self.rng = np.random.RandomState(seed)
        self.suggest_limit = suggest_limit
        self.budget_limit = budget_limit
        self.cost_limit = cost_limit
        self.total_time_limit = total_time_limit
        self.learner_time_limit = learner_time_limit
        self.learner_time_recoder = 0
        self.time_limit_recoder = 0
        self.total_time_recoder = 0
        self.suggest_counter = 0
        self.budget_recoder = 0
        self.cost_recoder = 0
        self.objective_function = objective_function

    def fix_boundary(self, individual):
        if self.fix_type == 'random':
            return np.where(
                (individual >= self.bounds.lb) & (individual <= self.bounds.ub),
                individual,
                self.rng.uniform(self.bounds.lb, self.bounds.ub,
                                 self.dimension))  # FIXME
        elif self.fix_type == 'clip':
            return np.clip(individual, self.bounds.lb, self.bounds.ub)

    def suggest(self, n_suggestions=1):
        st = time.time()
        ret = self._suggest(n_suggestions)
        self.total_time_recoder += time.time() - st
        self.suggest_counter += 1
        return ret

    @abstractmethod
    def _suggest(self, n_suggestions):  # output [meta param]
        """Get a suggestion from the optimizer.

        Parameters
        ----------
        n_suggestions : int
            Desired number of parallel suggestions in the output

        Returns
        -------
        next_guess : list of dict
            List of `n_suggestions` suggestions to evaluate the objective
            function. Each suggestion is a dictionary where each key
            corresponds to a parameter being optimized.
        """
        x_guess_configs = self.space.sample_configuration(size=n_suggestions)
        x_guess = [
            x_guess_config.get_dictionary()
            for x_guess_config in x_guess_configs
        ]
        return x_guess

    def _observe(self, trial_list: Trials):  # input [meta param]
        """Send an observation of a suggestion back to the optimizer.

        Parameters
        ----------
        features : list of features
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        pass

    def observe(self, trial_list: Trials):
        learner_train_time = 0
        for trial in trial_list:
            job_info = trial.info
            learner_train_time += job_info.get(Key.EVAL_TIME, 0)
            self.cost_recoder += job_info.get(Key.COST, 0)
            self.budget_recoder += job_info.get(Key.BUDGET, 0)

        st = time.time()
        ret = self._observe(trial_list)
        self.total_time_recoder += time.time() - st + learner_train_time
        self.learner_time_recoder += learner_train_time
        return ret

    def check_stop(self, ):
        if self.learner_time_recoder >= self.learner_time_limit or self.total_time_recoder >= self.total_time_limit or self.suggest_counter >= self.suggest_limit or self.budget_recoder >= self.budget_limit:
            return True
        else:
            return False

    def optimize(self):
        assert self.objective_function is not None
        while not self.check_stop():
            trial_list = self.suggest()
            for trial in (trial_list):
                print('Current suggest count={}.'.format(self.suggest_counter))
                info = trial.info.copy()
                res = self.objective_function(trial.config_dict, **info)
                if not isinstance(res, dict):
                    res = {Key.FUNC_VALUE: res}
                info.update(res)
                trial.add_observe_value(observe_value=info[Key.FUNC_VALUE],
                                        obs_info=info)
            self.observe(trial_list)



import logging

logging.basicConfig(level=logging.INFO)

import numpy as np
from ConfigSpace import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter

from xbbo.search_space.fast_example_problem import branin
# from xbbo.configspace.space import ConfigurationSpace
from xbbo.search_algorithm import alg_register


def run_one_exp(opt_name, max_call, seed):
    option_kwargs = {
        'turbo-1': {
            'name': 'turbo',
            'kwargs': {
                'num_tr': 1
            }
        },
        'turbo-2': {
            'name': 'turbo',
            'kwargs': {
                'num_tr': 2
            }
        },
        # 'rea':{
        #     'name':'rea',
        #     'kwargs':{
        #         'llambda': 100,
        #         # 'sample_size': 30
        #     }
        # }
    }
    # Build Configuration Space which defines all parameters and their ranges
    cs = ConfigurationSpace(seed=seed)
    x1 = UniformFloatHyperparameter("x1", -5, 10, default_value=0)
    x2 = UniformFloatHyperparameter("x2", 0, 15, default_value=0)
    cs.add_hyperparameters([x1, x2])
    if opt_name in option_kwargs:
        dic = option_kwargs[opt_name]
        hpopt = alg_register[dic['name']](space=cs,
                                          seed=seed,
                                          suggest_limit=max_call,
                                          **dic['kwargs'])
    else:
        hpopt = alg_register[opt_name](
            space=cs,
            seed=seed,
            suggest_limit=max_call,
        )

    # ---- Begin BO-loop ----
    for i in range(max_call):
        # suggest
        trial_list = hpopt.suggest()
        # evaluate
        for trial in trial_list:
            value = branin(trial.config_dict)
            # observe
            trial.add_observe_value(observe_value=value)
        hpopt.observe(trial_list=trial_list)

        print(value)
    losses = np.array(hpopt.trials._his_observe_value)
    # res = {'optimal': losses.min(), 'opt_call': losses.argmin() + 1}
    return losses  # np.minimum.accumulate(hpopt.trials.get_history()[0])


def benchmark(test_algs,
              func_to_call,
              max_call=200,
              repeat_num=10,
              father_seed=42,
              desc=''):
    import prettytable as pt

    results_all = []
    for test_alg in test_algs:
        rng = np.random.RandomState(father_seed)
        results_ = []
        for _ in range(repeat_num):
            seed = rng.randint(1e5)
            losses = func_to_call(test_alg, max_call, seed)
            res = [losses.min(), losses.argmin() + 1]

            results_.append(res)
        results_all.append(results_)
    results_all = np.array(results_all)  # (alg, repeat, 2)
    print(results_all)

    tb = pt.PrettyTable([
        "Method", "Minimum", "Best minimum", "Mean f_calls to min",
        "Std f_calls to min", "Fastest f_calls to min"
    ])
    # tb = pt.PrettyTable()
    # tb.field_names = []
    for i, results in enumerate(results_all):
        dict_result = {
            "mean": np.round(results.mean(axis=0), 3),
            "std": np.round(results.std(axis=0), 3),
            "best": np.round(results.min(axis=0), 3)
        }
        tb.add_row([
            '{}({})'.format(desc, test_algs[i]),
            '{:.3f}+/-{:.3f}'.format(dict_result["mean"][0],
                                     dict_result['std'][0]),
            dict_result["best"][0], dict_result["mean"][1],
            dict_result["std"][1],
            int(dict_result["best"][1])
        ])

    print(tb)
    return results_all


if __name__ == "__main__":
    # bore currently has some bugs
    test_algs = [
        'anneal', 'basic-bo', 'tpe', 'cem', 'cma-es', 'de', 'rs', 'rea',
        'turbo-1',
        'turbo-2',
        'bore'
    ]  # 'nsga2','bo-transfer','pbt'
    # benchmark(test_algs, run_one_exp, 200, 10, 42, desc='XBBO')
    benchmark(test_algs, run_one_exp, 20, 1, 42, desc='XBBO') # for fast test


import abc
import csv
from curses import A_ATTRIBUTES
import typing
import numpy as np
from xbbo.core.trials import Trials

# from xbbo.surrogate.base import Surrogate
from xbbo.surrogate.gaussian_process import GPR_sklearn
from xbbo.surrogate.transfer.tst import BaseModel
from xbbo.utils.constants import VERY_SMALL_NUMBER

class ABCWeightStategy(metaclass=abc.ABCMeta):
    def __init__(self,
                 cs,
                 base_models,
                 target_model,
                 rng,**kwargs):
        self.space = cs
        self.base_models = base_models
        self.target_model = target_model
        self.rng = rng
    @abc.abstractclassmethod
    def get_weight(self, trials: Trials):
        raise NotImplementedError()




class RankingWeight(ABCWeightStategy):
    def __init__(self,
                 cs,
                 base_models,
                 target_model,
                 rng,
                 budget,
                 rank_sample_num=256,
                 is_purn=False,alpha=0, **kwargs):
        super().__init__(cs, base_models, target_model, rng, **kwargs)
        self.rank_sample_num = rank_sample_num
        # self.iter = 0
        self.budget = budget
        self.is_purn = is_purn
        self.alpha = alpha


    def _compute_ranking_loss(self, f_samples, f_target):
        '''
        f_samples 'n_samples x (n) x n' -dim
        '''
        if f_samples.ndim == 3:  # for target model
            rank_loss = ((f_samples.diagonal(axis1=-2, axis2=-1)[:, :, None] <
                          f_samples) ^
                         (np.expand_dims(f_target, axis=-1) < np.expand_dims(
                             f_target, axis=-2))).sum(axis=-1).sum(axis=-1)
        else:
            rank_loss = ((np.expand_dims(f_samples, axis=-1) < np.expand_dims(
                f_samples, axis=-2)) ^
                         (np.expand_dims(f_target, axis=-1) < np.expand_dims(
                             f_target, axis=-2))).sum(axis=-1).sum(axis=-1)
        return rank_loss

    def _get_min_index(self, array):
        best_model_idxs = np.zeros(array.shape[1], dtype=np.int64)
        is_best_model = (array == array.min(axis=0))
        # idxs = np.argwhere(is_best_model)
        # mod, samp = idxs[:,0], idxs[:, 1]
        # self.rng.randint(np.bincount())
        for i in range(array.shape[1]):
            if is_best_model[-1, i]:
                best_model_idxs[i] = len(self.base_models)
            else:
                best_model_idxs[i] = self.rng.choice(
                    np.argwhere(is_best_model[:, i]).ravel().tolist())
        return best_model_idxs

    def get_weight(self, trials: Trials):
        t_x = trials.get_array()
        t_y = np.asarray(trials._his_observe_value)
        self.try_num = len(t_y)
        ranking_losses = []
        for base_model in self.base_models:
            mean, cov = base_model._predict(t_x, "full_cov")
            ranking_losses.append(
                self._compute_ranking_loss(
                    self.rng.multivariate_normal(mean,
                                                 cov,
                                                 size=self.rank_sample_num),
                    t_y))
        ranking_losses.append(
            self._compute_ranking_loss(self._get_loocv_preds(t_x, t_y), t_y))
        ranking_losses = np.array(ranking_losses)
        ranking_losses = self._delete_noise_knowledge(ranking_losses)
        # TODO argmin 多个最小处理
        # best_model_idxs = ranking_losses.argmin(axis=0)  # per sample都有一个best idx
        best_model_idxs = self._get_min_index(ranking_losses)

        # rank_weight = np.bincount(best_model_idxs, minlength=ranking_losses.shape[0]) / self.rank_sample_num
        rank_weight = np.bincount(best_model_idxs,
                                  minlength=ranking_losses.shape[0])

        return rank_weight  #/ rank_weight.sum()

    def _get_loocv_preds(self, x, y):
        masks = ~np.eye(self.try_num, dtype=np.bool_)
        x_cv = [x[m] for m in masks]
        y_cv = [y[m] for m in masks]
        samples = []
        for i in range(self.try_num):
            # kernel = self.new_gp.kernel.copy()
            model = GPR_sklearn(self.space,
                                rng=self.rng,
                                kernel=self.target_model.kernel,
                                do_optimize=False)
            model.train(x_cv[i], y_cv[i][:, None])
            mean, cov = model.predict(x, "full_cov")  # TODO 使用kernel子块
            # model = GaussianProcessRegressor(self.dim)
            # model.fit(x_cv[i], y_cv[i][:, None])
            # mean, cov = model.predict_with_cov(x)
            samples.append(
                self.rng.multivariate_normal(np.squeeze(mean), cov,
                                             self.rank_sample_num))
        return np.stack(samples, axis=1)

    # def predict_with_sigma(self, newX):
    #     models = [self.gps[d].cached_predict_with_sigma(newX) for d in range(self.old_D_num)]
    #     models.append(self.new_gp.predict_with_sigma(newX))
    #     models = np.asarray(models)
    #     mu = self.rank_weight.dot(models[:, 0])
    #     sigma = (self.rank_weight ** 2).dot(models[:, 1] ** 2)

    #     return mu, np.sqrt(sigma)

    def __delete_noise_knowledge(self, ranking_losses):
        if self.is_purn:
            # p_drop = 1 - (1 - self.x.shape[0] / 30) * (ranking_losses[:-1, :] < ranking_losses[-1, :]).sum(axis=-1) / (
            # self.rank_sample_num * (1 + self.alpha))
            # mask_to_remove = self.rng.binomial(1, p_drop) == 1

            target_loss = ranking_losses[-1]
            threshold = np.percentile(target_loss, 95)
            mask_to_remove = np.median(
                ranking_losses,
                axis=1) > threshold  # 中位数样本loss 比 95% 的target model结果差
            idx_to_remove = np.argwhere(mask_to_remove).ravel()
            ranking_losses = np.delete(ranking_losses, idx_to_remove, axis=0)
            # ranking_losses[idx_to_remove] = self.rank_sample_num
            for idx in reversed(idx_to_remove):  # TODO Remove permanent
                self.base_models.pop(idx)
            # self.old_D_num -= len(idx_to_remove)
        return ranking_losses
    
    def _delete_noise_knowledge(self, ranking_losses):
        if self.is_purn:
            p_drop = 1 - (1 - self.try_num / self.budget) * (ranking_losses[:-1, :] < ranking_losses[-1, :]).sum(axis=-1) / (
            self.rank_sample_num * (1 + self.alpha))
            mask_to_remove = self.rng.binomial(1, p_drop) == 1
            idx_to_remove = np.argwhere(mask_to_remove).ravel()
            # ranking_losses = np.delete(ranking_losses, idx_to_remove, axis=0)
            ranking_losses[idx_to_remove] = self.rank_sample_num
            # ranking_losses[idx_to_remove] = self.rank_sample_num
            # for idx in reversed(idx_to_remove):  # TODO Remove permanent
            #     self.base_models.pop(idx)
            # self.old_D_num -= len(idx_to_remove)
        return ranking_losses

class KernelRegress(ABCWeightStategy):
    def __init__(self, cs, base_models, target_model, rng, bandwidth=0.1,**kwargs):
        super().__init__(cs, base_models, target_model, rng, **kwargs)

        self.bandwidth = bandwidth
        # self.is_purn = is_purn

    def get_weight(self, trials: Trials):
        base_model_means = []
        for model in self.base_models:
            base_model_means.append(
                model._predict_normalize(trials.get_sparse_array(), None)[0])
        if not base_model_means:
            return []
        base_model_means = np.stack(base_model_means)  # [model, obs_num, 1]
        weight = self._naiveVersion(
            base_model_means, np.asarray(trials._his_observe_value))

        return weight  #/ weight.sum()

    def _kendallTauCorrelation(self, base_model_means, y):
        if y is None or len(y) < 2:
            return np.full(base_model_means.shape[0], 1)
        rank_loss = (
            (base_model_means[..., None] < base_model_means[..., None, :]) ^
            (y[..., None] < y[..., None, :])).astype('float')
        base_num, obs_num, _ = rank_loss.shape
        # rank_loss = rank_loss
        # rank_loss = (rank_loss < 1) * (1 - rank_loss**2) * 3 / 4
        num = obs_num * (obs_num - 1) // 2
        l = np.empty((base_num, num))
        idxs = np.triu_indices(obs_num, 1)
        for n in range(base_num):
            l[n] = rank_loss[n][idxs]
        l = np.linalg.norm(
            l / num/2,
            axis=1) / self.bandwidth  # meta feature : |source-target| / b
        l = (l < 1) * (1 - l * l) * 3 / 4
        
        return np.append(l, 3/4)
        # t = rank_loss.mean(axis=(-1, -2)) / self.bandwidth
        # return (t < 1) * (1 - t * t) * 3 / 4
        # return self.rho * (1 - t * t) if t < 1 else 0
    def _naiveVersion(self, base_model_means, y):
        meta_features = []
        for model_predit in base_model_means:
            meta_feature = []
            for i in range(len(y)):
                for j in range(i):
                    meta_feature.append((model_predit[i] > model_predit[j]) /
                                        (len(y) * (len(y) - 1)))
            meta_features.append(meta_feature)
        meta_feature = []
        for i in range(len(y)):
            for j in range(i):
                meta_feature.append((y[i] > y[j]) /
                                    (len(y) * (len(y) - 1)))
        meta_features.append(meta_feature)
        meta_features = np.asarray(meta_features)
        
        def kern(a, b, rho):
            def gamma(x):
                gamma = 3 / 4 * (1 - x ** 2) if x <= 1 else 0.0
                return gamma

            kern = gamma(np.linalg.norm(a - b) / rho)
            return kern
        weights = []
        for meta_feature in meta_features:
            weights.append(kern(meta_feature, meta_features[-1], self.bandwidth))
        return np.asarray(weights)
        
# class ProductExpert():
#     def __init__(self, cs, base_models, target_model, rng):
#         self.space = cs
#         self.base_models = base_models
#         self.target_model = target_model
#         # self.rank_sample_num = rank_sample_num
#         # self.iter = 0
#         self.rng = rng
#         # self.is_purn = is_purn

#     def get_weight(self, trials: Trials):
#         base_model_vars = []
#         for model in self.base_models:
#             base_model_vars.append(
#                 model.predict(trials.get_sparse_array())[1])
#         if not base_model_vars:
#             return []
#         base_model_vars = np.concatenate(base_model_vars)  # [model, obs_num, 1]
#         weight = self._naiveVersion(
#             base_model_vars, self.target_model.target_model_predict(trials.get_sparse_array())[1])

#         return weight  #/ weight.sum()


#     def _naiveVersion(self, base_model_vars, target_model_var):
#         beta = 1. / (len(base_model_vars)+1)
#         weights = [beta / var for var in base_model_vars]
#         weights.append(beta / target_model_var)
#         return np.array(weights)
        
class ZeroWeight(ABCWeightStategy):
    def __init__(self, cs, base_models, target_model, rng, **kwargs):
        super().__init__(cs, base_models, target_model, rng, **kwargs)

    def get_weight(self, trials: Trials):
        weight = np.zeros(len(self.base_models)+1)
        weight[-1] = 1

        return weight  #/ weight.sum()

        



# ref: https://github.com/thomas-young-2013/open-box/blob/master/openbox/surrogate/base/rf_with_instances_sklearn.py

import typing, logging
import numpy as np
from sklearn.ensemble import RandomForestRegressor
import threading
from joblib import Parallel, delayed
from sklearn.utils.fixes import _joblib_parallel_args
from sklearn.utils.validation import check_is_fitted
from xbbo.configspace.space import DenseConfigurationSpace
from xbbo.utils.util import get_types
try:
    from sklearn.ensemble.base import _partition_estimators
    old_sk_version = True
except:
    from sklearn.ensemble._base import _partition_estimators
    old_sk_version = False

from xbbo.surrogate.base import BaseRF

logger = logging.getLogger(__name__)


def _collect_prediction(predict, X, out, lock):
    """
    This is a utility function for joblib's Parallel.

    It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there.
    """
    prediction = predict(X, check_input=False)
    with lock:
        out.append(prediction)

class skRandomForestWithInstances(BaseRF):

    """Random forest that takes instance features into account.

    implement based on sklearn.ensemble.RandomForestRegressor

    Attributes
    ----------
    n_points_per_tree : int
    rf : RandomForestRegressor
        Only available after training
    unlog_y: bool
    seed : int
    types : np.ndarray
    bounds : list
    rng : np.random.RandomState
    logger : logging.logger
    """

    def __init__(self, configspace: DenseConfigurationSpace,
                 log_y: bool=False,
                 num_trees: int=10,
                 do_bootstrapping: bool=True,
                 n_points_per_tree: int=-1,
                 ratio_features: float=5. / 6.,
                 min_samples_split: int=3,
                 min_samples_leaf: int=3,
                 max_depth: int=2**20,
                 eps_purity: float=1e-8,
                 max_num_nodes: int=2**20,
                 rng: np.random.RandomState = np.random.RandomState(42),
                 n_jobs: int=None,
                 **kwargs):
        """
        Parameters
        ----------
        types : np.ndarray (D)
            Specifies the number of categorical values of an input dimension where
            the i-th entry corresponds to the i-th input dimension. Let's say we
            have 2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass np.array([2, 0]). Note that we count starting from 0.
        bounds : list
            Specifies the bounds for continuous features.
        log_y: bool
            y values (passed to this RF) are expected to be log(y) transformed;
            this will be considered during predicting
        num_trees : int
            The number of trees in the random forest.
        do_bootstrapping : bool
            Turns on / off bootstrapping in the random forest.
        n_points_per_tree : int
            Number of points per tree. If <= 0 X.shape[0] will be used
            in _train(X, y) instead
        ratio_features : float
            The ratio of features that are considered for splitting.
        min_samples_split : int
            The minimum number of data points to perform a split.
        min_samples_leaf : int
            The minimum number of data points in a leaf.
        max_depth : int
            The maximum depth of a single tree.
        eps_purity : float
            The minimum difference between two target values to be considered
            different
        max_num_nodes : int
            The maxmimum total number of nodes in a tree
        seed : int
            The seed that is passed to the random_forest_run library.
        n_jobs : int, default=None
            The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
            :meth:`decision_path` and :meth:`apply` are all parallelized over the
            trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context. ``-1`` means using all processors. See :term:`Glossary
            <n_jobs>` for more details.
        """
        types, bounds = get_types(configspace)
        super().__init__(configspace, types, bounds, **kwargs)


        self.log_y = log_y
        if self.log_y:
            raise NotImplementedError
        self.rng = rng

        self.num_trees = num_trees
        self.do_bootstrapping = do_bootstrapping
        max_features = None if ratio_features > 1.0 else \
            int(max(1, types.shape[0] * ratio_features))
        self.max_features = max_features
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_depth = max_depth
        self.epsilon_purity = eps_purity
        self.max_num_nodes = max_num_nodes

        self.n_points_per_tree = n_points_per_tree
        self.n_jobs = n_jobs

        self.rf = None  # type: RandomForestRegressor

    def _train(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """Trains the random forest on X and y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, ]
            The corresponding target values.

        Returns
        -------
        self
        """
        X = self._impute_inactive(X)
        self.X = X
        self.y = y.flatten()

        if self.n_points_per_tree <= 0:
            self.num_data_points_per_tree = self.X.shape[0]
        else:
            self.num_data_points_per_tree = self.n_points_per_tree
        if old_sk_version:
            self.rf = RandomForestRegressor(
                n_estimators=self.num_trees,
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                min_samples_leaf=self.min_samples_leaf,
                max_features=self.max_features,
                # max_samples=self.num_data_points_per_tree,
                max_leaf_nodes=self.max_num_nodes,
                min_impurity_decrease=self.epsilon_purity,
                bootstrap=self.do_bootstrapping,
                n_jobs=self.n_jobs,
                random_state=self.rng,
            )
        else:
            self.rf = RandomForestRegressor(
                n_estimators=self.num_trees,
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                min_samples_leaf=self.min_samples_leaf,
                max_features=self.max_features,
                max_samples=self.num_data_points_per_tree,
                max_leaf_nodes=self.max_num_nodes,
                min_impurity_decrease=self.epsilon_purity,
                bootstrap=self.do_bootstrapping,
                n_jobs=self.n_jobs,
                random_state=self.rng,
            )
        self.rf.fit(self.X, self.y)
        return self

    def predict_mean_var(self, X: np.ndarray):
        if old_sk_version:
            check_is_fitted(self.rf, 'estimators_')
        else:
            check_is_fitted(self.rf)
        # Check data
        if X.ndim == 1:
            X = X.reshape((1, -1))
        X = self.rf._validate_X_predict(X)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.rf.n_estimators, self.rf.n_jobs)

        # collect the output of every estimator
        all_y_preds = list()

        # Parallel loop
        lock = threading.Lock()
        Parallel(n_jobs=n_jobs, verbose=self.rf.verbose,
                 **_joblib_parallel_args(require="sharedmem"))(
            delayed(_collect_prediction)(e.predict, X, all_y_preds, lock)
            for e in self.rf.estimators_)
        all_y_preds = np.asarray(all_y_preds, dtype=np.float64)

        m = np.mean(all_y_preds, axis=0)
        v = np.var(all_y_preds, axis=0)
        return m, v

    def _predict(self, X: np.ndarray, **kwargs) -> typing.Tuple[np.ndarray, np.ndarray]:
        """Predict means and variances for given X.

        Parameters
        ----------
        X : np.ndarray of shape = [n_samples,
                                   n_features (config + instance features)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, 1]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, 1]
            Predictive variance
        """
        if len(X.shape) != 2:
            raise ValueError(
                'Expected 2d array, got %dd array!' % len(X.shape))
        if X.shape[1] != self.types.shape[0]:
            raise ValueError('Rows in X should have %d entries but have %d!' % (self.types.shape[0], X.shape[1]))
        X = self._impute_inactive(X)
        if self.log_y:
            raise NotImplementedError
        else:
            means, vars_ = self.predict_mean_var(X)

        return means.reshape((-1, 1)), vars_.reshape((-1, 1))

    def predict_marginalized_over_instances(self, X: np.ndarray):
        """Predict mean and variance marginalized over all instances.

        Returns the predictive mean and variance marginalised over all
        instances for a set of configurations.

        Note
        ----
        This method overwrites the same method of ~smac.epm.base_epm.AbstractEPM;
        the following method is random forest specific
        and follows the SMAC2 implementation;
        it requires no distribution assumption
        to marginalize the uncertainty estimates

        Parameters
        ----------
        X : np.ndarray
            [n_samples, n_features (config)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, 1]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, 1]
            Predictive variance
        """

        if self.log_y:
            raise NotImplementedError
        else:
            return super().predict_marginalized_over_instances(X)



'''
Reference: https://github.com/automl/DEHB
'''

from typing import List
import numpy as np

# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trials, Trial
from xbbo.search_algorithm.multi_fidelity.utils.bracket_manager import BasicConfigGenerator, ConfigGenerator, SHBracketManager
from xbbo.search_algorithm.random_optimizer import RandomOptimizer

from xbbo.utils.constants import Key
from .. import alg_register

class HB_CG(BasicConfigGenerator, RandomOptimizer):
    def __init__(self, cs, budget, max_pop_size, rng, **kwargs) -> None:
        BasicConfigGenerator.__init__(self, cs, budget, max_pop_size, rng, **kwargs)
        RandomOptimizer.__init__(self, space=cs, init_budget=0, **kwargs)
        self.reset(max_pop_size)

alg_marker = 'hb'


@alg_register.register('hb')
class HB(AbstractOptimizer):
    name = alg_marker
    def __init__(self,
                 space: DenseConfigurationSpace,
                 budget_bound=[9, 729],
                 eta: int = 3,
                 seed: int = 42,
                 round_limit: int = 1,
                 bracket_limit: int = np.inf,
                 boundary_fix_type='random',
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                #    encoding_cat='bin',
                                #    encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)

        # Uniform2Gaussian.__init__(self,)
        self.dimension = self.space.get_dimensions()
        self.bounds = self.space.get_bounds()
        self.fix_type = boundary_fix_type
        # self.dimension = len(configs)
        # self.candidates = [None] * self.llambda
        self.trials = Trials(dim=self.dimension)
        self.current_best = None
        self.current_best_fitness = np.inf

        self.min_budget, self.max_budget = budget_bound
        self.eta = eta
        self.max_SH_iter = None
        self.budgets = None
        if self.min_budget is not None and \
           self.max_budget is not None and \
           self.eta is not None:
            self.max_SH_iter = -int(
                np.log(self.min_budget / self.max_budget) /
                np.log(self.eta)) + 1  # max_sh_iter是数量，0~t,(t+1)个
            self.budgets: List[int] = self.max_budget * np.power(
                self.eta, -np.linspace(
                    start=self.max_SH_iter - 1, stop=0, num=self.max_SH_iter))
        self._max_pop_size = None
        self.active_brackets = []  # list of SHBracketManager objects

        self.round_limit = round_limit
        self.bracket_limit = bracket_limit
        self.bracket_counter = -1 # completed bracket count
        self.round_recoder = -1 # completed cound count
        self.learner_time_recoder = 0
        self.total_time_recoder = 0
        self.kwargs = kwargs
        self._get_max_pop_sizes()
        self._init_subpop(**kwargs)

    def check_stop(self, ):
        if AbstractOptimizer.check_stop(self) or self.round_recoder >= self.round_limit or self.bracket_counter >= self.bracket_limit : # FIXME
            return True
        else:
            return False

    def _start_new_bracket(self):
        """ Starts a new bracket based on Hyperband
        """
        # start new bracket
        # self.round_recoder = self.bracket_counter // self.max_SH_iter
        self.bracket_counter += 1  # iteration counter gives the bracket count or bracket ID
        n_configs, budgets = self._get_next_bracket_space(self.bracket_counter)
        bracket = SHBracketManager(n_configs=n_configs,
                                   budgets=budgets,
                                   bracket_id=self.bracket_counter)
        self.active_brackets.append(bracket)
        return bracket

    def _get_next_bracket_space(self, iteration):
        '''Computes the Successive Halving spacing

        Given the iteration index, computes the budget spacing to be used and
        the number of configurations to be used for the SH iterations.

        Parameters
        ----------
        iteration : int
            Iteration index
        clip : int, {1, 2, 3, ..., None}
            If not None, clips the minimum number of configurations to 'clip'

        Returns
        -------
        ns : array
        budgets : array
        '''
        # number of 'SH runs'
        s = self.max_SH_iter - 1 - (iteration % self.max_SH_iter)
        # budget spacing for this iteration
        budgets = self.budgets[(-s - 1):]
        # number of configurations in that bracket
        n0 = int(np.floor((self.max_SH_iter) / (s + 1)) * self.eta**s)
        ns = [max(int(n0 * (self.eta**(-i))), 1) for i in range(s + 1)]
        # if self.min_clip is not None and self.max_clip is not None:
        #     ns = np.clip(ns, a_min=self.min_clip, a_max=self.max_clip)
        # elif self.min_clip is not None:
        #     ns = np.clip(ns, a_min=self.min_clip, a_max=np.max(ns))

        return ns, budgets

    def _suggest(self, n_suggestions=1):
        trial_list = []
        for n in range(n_suggestions):
            if len(self.active_brackets) == 0 or \
                np.all([bracket.is_bracket_done() for bracket in self.active_brackets]):
                bracket = self._start_new_bracket()
            else:
                for _bracket in self.active_brackets:
                    # check if _bracket is not waiting for previous rung results of same bracket
                    # _bracket is not waiting on the last rung results
                    # these 2 checks allow DEHB to have a "synchronous" Successive Halving
                    if not _bracket.previous_rung_waits(
                    ) and _bracket.is_pending():
                        # bracket eligible for job scheduling
                        bracket = _bracket
                        break
                if bracket is None:
                    # start new bracket when existing list has all waiting brackets
                    bracket = self._start_new_bracket()  # new column

                    # budget that the SH bracket allots
            budget = bracket.get_next_job_budget()
            # new individual
            candidate, parent_id = self._acquire_candidate(bracket, budget)
            # notifies the Bracket Manager that a single config is to run for the budget chosen
            # job_info = {
            #     "config": config,
            #     Key.BUDGET: budget,
            #     "parent_id": parent_id,
            #     "bracket_id": bracket.bracket_id
            # }
            # # return job_info

            config = DenseConfiguration.from_array(self.space, candidate)
            trial_list.append(
                Trial(config,
                      config_dict=config.get_dictionary(),
                      array=candidate,
                      info={
                          Key.BUDGET: budget,
                          "parent_id": parent_id,
                          "bracket_id": bracket.bracket_id
                      },
                      origin=self.name))
        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, True)
            fitness = trial.observe_value
            job_info = trial.info
            # learner_train_time = job_info.get(Key.EVAL_TIME, 0)
            budget = job_info[Key.BUDGET]
            parent_id = job_info['parent_id']
            individual = trial.array  # TODO
            for bracket in self.active_brackets:
                if bracket.bracket_id == job_info['bracket_id']:
                    # registering is IMPORTANT for Bracket Manager to perform SH
                    bracket.register_job(budget)  # may be new row
                    # bracket job complete
                    bracket.complete_job(
                        budget)  # IMPORTANT to perform synchronous SH
            self.cg[budget].population[parent_id] = individual
            self.cg[budget].population_fitness[parent_id] = fitness
            # updating incumbents
            if fitness < self.current_best_fitness:
                self.current_best = individual
                self.current_best_fitness = trial.observe_value
                self.current_best_trial = trial
        self.cg[budget]._observe(trial_list)


        self._clean_inactive_brackets()
        # if len(self.active_brackets) == 0: # complete current bracket
        #     self.round_recoder = self.bracket_counter // self.max_SH_iter
            # self._start_new_bracket()

    def _clean_inactive_brackets(self):
        """ Removes brackets from the active list if it is done as communicated by Bracket Manager
        """
        self.active_brackets = [
            bracket for bracket in self.active_brackets
            if ~bracket.is_bracket_done()
        ]
        if len(self.active_brackets) == 0:
            self._start_new_bracket()
            self.round_recoder = self.bracket_counter // self.max_SH_iter
        return

    def _get_max_pop_sizes(self):
        """Determines maximum pop size(config num) for each budget
        """
        self._max_pop_size = {}
        for i in range(self.max_SH_iter):
            n, r = self._get_next_bracket_space(i)
            for j, r_j in enumerate(r):
                self._max_pop_size[r_j] = max(
                    n[j], self._max_pop_size[r_j]
                ) if r_j in self._max_pop_size.keys() else n[j]

    def _init_subpop(self, **kwargs):
        """ List of DE objects corresponding to the budgets (fidelities)
        """
        self.cg = {}
        for i, b in enumerate(self._max_pop_size.keys()):
            self.cg[b] = HB_CG(self.space,
                                         budget=b,
                                         max_pop_size=self._max_pop_size[b],
                                         rng=self.rng,
                                         **kwargs)
            # self.cg[b].population = self.cg[b].init_population(
            #     pop_size=self._max_pop_size[b])
            # self.cg[b].population_fitness = np.array([np.inf] * self._max_pop_size[b])
            # # adding attributes to DEHB objects to allow communication across subpopulations
            # self.cg[b].parent_idx = 0
            # self.cg[b].promotion_pop = None
            # self.cg[b].promotion_fitness = None

    def _acquire_candidate(self, bracket, budget):
        """ Generates/chooses a configuration based on the budget and iteration number
        """
        parent_id = self._get_next_idx_for_subpop(budget, bracket)

        if budget != bracket.budgets[0]:
            if bracket.is_new_rung():
                # TODO: check if generalizes to all budget spacings
                lower_budget, num_configs = bracket.get_lower_budget_promotions(
                    budget)
                self._get_promotion_candidate(lower_budget, budget,
                                              num_configs)
            # else: # 每一列中的第一行，随机生成config
            target = self.cg[budget].population[parent_id]

        else: # config generator
            trial = self.cg[budget]._suggest()[0]

            target = trial.array
        # parent_id = self._get_next_idx_for_subpop(budget, bracket)

        # target = self.fix_boundary(target)
        return target, parent_id

    def _get_promotion_candidate(self, low_budget, high_budget, n_configs):
        """ Manages the population to be promoted from the lower to the higher budget.

        This is triggered or in action only during the first full HB bracket, which is equivalent
        to the number of brackets <= max_SH_iter.
        """
        # finding the individuals that have been evaluated (fitness < np.inf)
        evaluated_configs = np.where(
            self.cg[low_budget].population_fitness != np.inf)[0]
        promotion_candidate_pop = self.cg[low_budget].population[
            evaluated_configs]
        promotion_candidate_fitness = self.cg[low_budget].population_fitness[
            evaluated_configs]
        # ordering the evaluated individuals based on their fitness values
        pop_idx = np.argsort(promotion_candidate_fitness)
        # n_configs = len(self.cg[high_budget].population) # not sure eta multiple configs
        elite_index = pop_idx[:n_configs]
        # self.cg[high_budget].population_fitness = self.cg[low_budget].population_fitness[elite_index]
        self.cg[high_budget].population[:len(elite_index)] = promotion_candidate_pop[
            elite_index]
        # self.cg[high_budget].population_fitness[len(elite_index):] = np.inf
        self.cg[high_budget].population_fitness[:] = np.inf
        # return

    def _get_next_idx_for_subpop(self, budget, bracket):
        """ Maintains a looping counter over a subpopulation, to iteratively select a parent
        """
        parent_id = self.cg[budget].parent_idx
        self.cg[budget].parent_idx += 1
        # self.cg[budget].parent_idx = self.cg[
        #     budget].parent_idx % self._max_pop_size[budget]
        self.cg[budget].parent_idx = self.cg[
            budget].parent_idx % bracket.current_n_config
        return parent_id


opt_class = HB



import numpy as np

class Category():
    '''
    every point has a instance of this class
    '''
    def __init__(self, choices, prior=None, prior_w=0):
        '''
        choice should be a int number
        '''
        self.choices = choices
        self.prior_w = prior_w
        if prior is None:
            self.prob = [1/self.choices for _ in range(self.choices)]
        else:
            self.prob = prior
        self._prior = self.prob

    def sample(self, p=None):
        return np.random.multinomial(1, self.prob if p is None else p, size=1)

    def log_pdf(self, x):
        return np.log(self.prob[x])

    def pdf(self, x):
        return self.prob[x]

    def random_sample(self):
        return np.random.choice(self.choices)

    def update(self, x, weights=None):
        counts = np.bincount(x, weights, minlength=self.choices)
        self.prob = (1 - self.prior_w) * counts + self.prior_w * self._prior


class Uniform():
    def __init__(self, low=0, high=1):
        self.low = low
        self.high = high

    def sample(self, low=None, high=None):
        if low is None:
            low = self.low
        if high is None:
            high = self.high
        return np.random.uniform(low, high)

    def random_sample(self):
        return np.random.uniform(self.low, self.high)


class Normal():
    def __init__(self, mu=0, sigma=1):
        self.mu = mu
        self.sigma = sigma

    def sample(self, mu=None, sigma=None):
        if mu is None:
            mu = self.mu
        if sigma is None:
            sigma = self.sigma
        return np.random.normal(mu, sigma)

    def random_sample(self,  ):
        mu = self.mu
        sigma = self.sigma
        return np.random.normal(mu, sigma)



import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from torchvision import datasets, transforms
from torch.utils.data import DataLoader, Dataset

from xbbo.core import TestFunction

class LossIsNaN(Exception):
    pass


class ConvNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=(5,5))
        self.conv2 = nn.Conv2d(10, 20, kernel_size=(5,5))
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)

class MnistShuffleTrainData(datasets.MNIST):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        np.random.seed(0)
        self.idx_shuffle = np.random.permutation(len(self))

    def __getitem__(self, item):
        return super().__getitem__(self.idx_shuffle[item])

class Model(TestFunction):
    trn_dataset = None
    tst_dataset = None
    def __init__(self, cfg, **kwargs):
        # np.random.seed(cfg.GENERAL.random_seed)
        self.cfg = cfg
        # self.dim = 30
        # assert self.dim % 2 == 0
        super().__init__()

        self.api_config = self._load_api_config()
        transform = transforms.Compose([transforms.ToTensor(),
                                        transforms.Normalize((0.1307,),
                                                             (0.3081,))])
        if Model.trn_dataset is None:
            Model.trn_dataset = MnistShuffleTrainData(kwargs['data_path'], train=True, download=True,
                                            transform=transform)
            Model.tst_dataset = datasets.MNIST(kwargs['data_path'], train=False, download=True,
                                                transform=transform)
        self.device = torch.device(kwargs.get('device', 'cpu'))
        self.train_loader = DataLoader(Model.trn_dataset, batch_size=64, shuffle=False)
        self.test_loader = DataLoader(Model.tst_dataset, batch_size=64, shuffle=False)
        self.net = ConvNet().to(self.device)
        # self.opt_wrap = lambda params: optim.SGD(self.net.parameters(), lr=lr, momentum=momentum)
        self.opt = optim.SGD(self.net.parameters(), lr=0.11, momentum=0.9)
        self.loss_fn = F.nll_loss
        self.step_num = 0
        self.ready = False # not ready
        self.history_hp = [] # for record strategy
        self.trajectory_hp = []
        self.trajectory_score = [] # 记录该个体score过程
        self.history_score = [] # 记录使用了（考虑权重迁移）hp-stategy后的score过程

    def __len__(self): # one epoch has how many batchs
        return len(self.train_loader)

    def update_hp(self, params: dict):
        self.history_hp.append((self.step_num, params)) # 在该steps上更改超参，acc为该step时的结果（受该step*前*所有超参影响）
        self.trajectory_hp.append((self.step_num, params))
        for hyperparam_name, v in params.items():
            for param_group in self.opt.param_groups:
                param_group[hyperparam_name] = v

    def step(self, num): # train need training(optimizer)
        self.net.train()
        st = self.step_num % len(self.train_loader)
        ed = st + num
        it = 0
        while it < ed:

            for (inp, target) in (self.train_loader):

                if it < st:
                    it += 1
                    continue
                # it += 1
                inp = inp.to(self.device)
                target = target.to(self.device)
                output = self.net(inp)
                loss = self.loss_fn(output, target)
                if np.isnan(loss.item()):
                    print("Loss is NaN.")
                    self.step_num += ed - it
                    it = ed
                    break
                    # raise LossIsNaN
                self.opt.zero_grad()
                loss.backward()
                self.opt.step()
                self.step_num += 1
                it += 1
                if ed == it:
                    break
        # inp, target = next(self.train_loader)


        # if self.step_num % len(Model.trn_dataset) == 0:
        #     self.ready = True



    def evaluate(self): # val no training need(optimizer)
        correct = 0
        self.net.eval()
        with torch.no_grad():
            for inp, target in self.test_loader:
                inp = inp.to(self.device)
                target = target.to(self.device)
                output = self.net(inp)
                correct += (output.max(1)[1] == target).sum().cpu().item()
        acc = 100 * correct / len(self.tst_dataset)
        self.score = -1 if np.isnan(acc) else acc
        self.trajectory_score.append((self.step_num, self.score))
        self.history_score.append((self.step_num, self.score))
        return self.score

    def load_checkpoint(self, checkpoint):
        self.net.load_state_dict(checkpoint['model_state_dict'])
        self.opt.load_state_dict(checkpoint['optim_state_dict'])

    def save_checkpoint(self):
        checkpoint = dict(model_state_dict=self.net.state_dict(),
                          optim_state_dict=self.opt.state_dict())
        return checkpoint

    # def save_checkpoint(self, checkpoint_path):
    #     checkpoint = dict(model_state_dict=self.model.state_dict(),
    #                       optim_state_dict=self.optimizer.state_dict())
    #     torch.save(checkpoint, checkpoint_path)

    # def load_checkpoint(self, checkpoint_path):
    #     checkpoint = torch.load(checkpoint_path)
    #     self.net.load_state_dict(checkpoint['model_state_dict'])
    #     self.opt.load_state_dict(checkpoint['optim_state_dict'])

    def _load_api_config(self):
        return {
            'lr': {
                'type': 'float', 'warp': 'log10', 'range': [1e-5, 1]},
            'momentum': {
                'type': 'float', 'warp': 'linear', 'range': [0.1, 0.999]
            }
        }





from typing import List
import numpy as np

# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.multi_fidelity.hyperband import HB
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trials, Trial
from xbbo.search_algorithm.multi_fidelity.utils.bracket_manager import RFHB_ConfigGenerator, SHBracketManager
from xbbo.utils.constants import Key
from .. import alg_register

alg_marker = 'rfhb'

@alg_register.register(alg_marker)
class RFHB(HB):
    name = alg_marker
    def __init__(
            self,
            space: DenseConfigurationSpace,
            budget_bound=[9, 729],
            #  mutation_factor=0.5,
            #  crossover_prob=0.5,
            #  strategy='rand1_bin',
            eta: int = 3,
            seed: int = 42,
            round_limit: int = 1,
            bracket_limit=np.inf,
            boundary_fix_type='random',
            **kwargs):
        HB.__init__(self,
                    space,
                    budget_bound,
                    eta,
                    seed=seed,
                    round_limit=round_limit,
                    bracket_limit=bracket_limit,
                    boundary_fix_type=boundary_fix_type,
                    **kwargs)

    def _init_subpop(self, **kwargs):
        """ List of DE objects corresponding to the budgets (fidelities)
        """
        self.cg = {}
        for i, b in enumerate(self._max_pop_size.keys()):
            self.cg[b] = RFHB_ConfigGenerator(
                self.space,
                budget=b,
                max_pop_size=self._max_pop_size[b],
                rng=self.rng,
                eta = self.eta,
                **kwargs)

    def _acquire_candidate(self, bracket, budget):
        """ Generates/chooses a configuration based on the budget and iteration number
        """
        if bracket.is_new_rung():
            lower_budget, num_configs = bracket.get_lower_budget_promotions(
                    budget)
            self.cg[budget].reset(bracket.current_n_config)
            self.cg[budget].population_fitness[:] = np.inf
            if budget != bracket.budgets[0]:  # 每一列中的第二行开始，进行seed
                # TODO: check if generalizes to all budget spacings
                
                self._get_promotion_candidate(lower_budget, budget,
                                              num_configs)
            else:
            # else: # 每一列中的第一行，随机生成config,或者使用随机森林拒绝采样
                self.cg[budget].update_new_subpopulation(self.cg[budget]._init_population(num_configs))

        parent_id = self._get_next_idx_for_subpop(budget, bracket)
        target = self.cg[budget].population[parent_id]

        target = self.fix_boundary(target)
        return target, parent_id

    def _get_promotion_candidate(self, low_budget, high_budget, n_configs):
        """ Manages the population to be promoted from the lower to the higher budget.

        This is triggered or in action only during the first full HB bracket, which is equivalent
        to the number of brackets <= max_SH_iter.
        """
        # finding the individuals that have been evaluated (fitness < np.inf)
        evaluated_configs = np.where(
            self.cg[low_budget].population_fitness != np.inf)[0]
        promotion_candidate_pop = self.cg[low_budget].population[
            evaluated_configs]
        promotion_candidate_fitness = self.cg[low_budget].population_fitness[
            evaluated_configs]
        # ordering the evaluated individuals based on their fitness values
        pop_idx = np.argsort(promotion_candidate_fitness)
        labels = promotion_candidate_fitness.copy()
        # labels = np.zeros_like(promotion_candidate_fitness)
        elite_index = pop_idx[:n_configs]

        # labels[elite_index] = 1
        # fiting low; gen high
        self.cg[low_budget].add_fitting(promotion_candidate_pop, labels)
        # promotion top to down
        self.cg[high_budget].update_new_subpopulation(
            self.cg[low_budget].population[elite_index])

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)
            fitness = trial.observe_value
            job_info = trial.info
            budget = job_info[Key.BUDGET]
            parent_id = job_info['parent_id']
            individual = trial.array  # TODO
            for bracket in self.active_brackets:
                if bracket.bracket_id == job_info['bracket_id']:
                    # registering is IMPORTANT for Bracket Manager to perform SH
                    bracket.register_job(budget)  # may be new row
                    # bracket job complete
                    bracket.complete_job(
                        budget)  # IMPORTANT to perform synchronous SH
            # carry out DE selection
            if fitness <= self.cg[budget].population_fitness[parent_id]:  # TODO
                self.cg[budget].population[parent_id] = individual
                self.cg[budget].population_fitness[parent_id] = fitness
                # updating incumbents
                if fitness < self.current_best_fitness:
                    self.current_best = individual
                    self.current_best_fitness = trial.observe_value
                    self.current_best_trial = trial

        self._clean_inactive_brackets()
        # if len(self.active_brackets) == 0:  # complete current bracket
        #     self.round_recoder = self.bracket_counter // self.max_SH_iter

opt_class = RFHB



import copy
import math
from typing import Optional, List, Tuple, cast

import numpy as np

from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace, convert_denseConfigurations_to_array
from xbbo.core.trials import Trials
from xbbo.initial_design import ALL_avaliable_design
from . import alg_register

@alg_register.register('pbt')
class PBT():
    def __init__(self,
                 space: DenseConfigurationSpace,
                 pop_size: int = 30,
                 initial_design: str = 'random',
                 seed: int = 42,
                 fraction: float = 0.2,
                 **kwargs):
        # FeatureSpace_gaussian.__init__(self, self.space.dtypes_idx_map)
        
        self.space = space
        self.rng = np.random.RandomState(seed)
        self.dimension = self.space.get_dimensions()
        self.pop_size = pop_size if pop_size else 4 + math.floor(3 * math.log(self.dimension))
        self.pop_size = pop_size
        self.init_budget = pop_size

        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, init_budget=self.init_budget)

        # self.population_hp = [{'h1':1.0, 'h2':0.0}, {'h1':0.0,'h2':1.0}]
        self.initial_design_configs = self.initial_design.select_configurations(
        )
        self.population_configs = self.initial_design_configs
        self.population_hp_array = convert_denseConfigurations_to_array(self.population_configs)
        self.trials = Trials(dim=self.dimension)

        # self.data_shuffle_seed = kwargs.get('seed', 0)
        self.fraction = fraction
        self.population_losses_his = []

    def init_model_hp(self, population_model):
        for i, model in enumerate(population_model):
            model.loss = model.evaluate()
            model.update_hp(self.population_configs[i].get_dictionary())
        self.population_losses_his.append(
            [model.loss
             for model in population_model])

    def exploit_and_explore(self, population_model, losses):
        self.population_losses_his.append(losses)
        s_id = np.argsort(losses)
        top_ids = s_id[:int(self.fraction * len(s_id))]
        bot_ids = s_id[-int(self.fraction * len(s_id)):]
        # nobot_ids = s_id[:-self.fraction*len(s_id)]
        for bot_id in bot_ids:
            # exploit
            top_id = self.rng.choice(top_ids)
            checkpoint = population_model[top_id].save_checkpoint()
            population_model[bot_id].load_checkpoint(checkpoint)
            self.population_hp_array[bot_id] = self.population_hp_array[
                top_id].copy()
            # explore
            self.population_hp_array[bot_id] = np.clip(
                self.population_hp_array[bot_id] +
                np.random.normal(0, 0.2, size=self.dimension), 0, 1)

            x_array = self.feature_to_array(self.population_hp_array[bot_id],
                                            self.sparse_dimension)
            x_unwarped = DenseConfiguration.array_to_dict(self.space, x_array)
            self.population_hp[bot_id] = x_unwarped
            population_model[bot_id].history_hp = copy.copy(
                population_model[top_id].history_hp)
            population_model[bot_id].history_loss = copy.copy(
                population_model[top_id].history_loss)
            population_model[bot_id].update_hp(x_unwarped)

    def exploit_and_explore_toy(self, population_model, losses):
        self.population_losses_his.append(losses)
        s_id = np.argsort(losses)
        top_num = max(int(self.fraction * len(s_id)), 1)
        top_ids = s_id[:top_num]
        bot_ids = s_id[-top_num:]
        # nobot_ids = s_id[:-self.fraction*len(s_id)]
        for bot_id in bot_ids:  # TODO Toy every point do explore
            # exploit
            top_id = np.random.choice(top_ids)
            checkpoint = population_model[top_id].save_checkpoint()
            population_model[bot_id].load_checkpoint(checkpoint)
            # self.population_hp_array[bot_id] = self.population_hp_array[top_id].copy() # TODO Toy
            # explore
            self.population_hp_array[bot_id] = np.clip(  # TODO Toy
                self.population_hp_array[top_id] +
                self.rng.normal(0, 0.1, size=self.dimension), 0, 1)

            
            new_config = DenseConfiguration.from_array(self.space, self.population_hp_array[bot_id])
            self.population_configs[bot_id] = new_config
            # self.population_hp[bot_id] = x_unwarped
            # population_model[bot_id].history_hp = copy.copy(
                # population_model[top_id].history_hp)
            population_model[bot_id].history_loss = copy.copy(
                population_model[top_id].history_loss)
            population_model[bot_id].update_hp(new_config.get_dictionary())


opt_class = PBT



import numpy as np
import json
from time import time
import tqdm

from xbbo.search_space import problem_register
from xbbo.search_algorithm import alg_register
from xbbo.configspace import build_space
from xbbo.utils.record import Record


class Transfer_BBO:

    def __init__(self, cfg):
        # setup TestProblem
        self.cfg = cfg
        self.function_instance = problem_register[cfg.TEST_PROBLEM.name](cfg)

        self.api_config = self.function_instance.get_api_config()  # 优化的hp
        self.config_spaces = build_space(self.api_config)


        # Setup optimizer
        opt_class = alg_register[cfg.OPTM.name]
        self.optimizer_instance = opt_class(self.config_spaces, **dict(cfg.OPTM.kwargs))

        old_D_x_params, old_D_y, new_D_x_param = self.function_instance.array_to_config(ret_param=True)
        # old_D_x_params, old_D_y, new_D_x_param = self.function_instance.array_to_config()
        # self.function_instance.cache(new_D_x_param)

        self.optimizer_instance.prepare(old_D_x_params, old_D_y, new_D_x_param,
                                        np.argsort(list(self.api_config.keys())), params=True)
        # self.optimizer_instance.prepare(old_D_x_params, old_D_y, new_D_x_param,
        #                                 np.argsort(list(self.api_config.keys())))
        # old_D_x_params, old_D_y, new_D_x_param = self.function_instance.array_to_config()
        # # self.function_instance.cache(new_D_x_param)
        #
        # self.optimizer_instance.prepare(old_D_x_params, old_D_y, new_D_x_param, np.argsort(list(self.api_config.keys())))

        self.n_suggestions = cfg.OPTM.n_suggestions
        self.n_obj = cfg.OPTM.n_obj

        assert self.n_suggestions >= 1, "batch size must be at least 1"
        assert self.n_obj >= 1, "Must be at least one objective"


        # self.suggest_time = np.zeros(n_calls)
        # self.observe_time = np.zeros(n_calls)
        # self.eval_time = np.zeros((n_calls, n_suggestions))
        # self.function_evals = np.zeros((n_calls, n_suggestions, self.n_obj))
        # self.suggest_log = [None] * n_calls
        self.n_calls = cfg.OPTM.max_call
        self.record = Record(self.cfg)

        # # 预加载观测
        # if history:
        #     self._load_history_and_obs(history)



    # def _load_history_and_obs(self, filename):
    #     obs_x, obs_y, isFeature = load_history(filename)
    #     for i in range(len(obs_x)):
    #         if not isFeature:
    #             obs_x[i] = Configurations.dictUnwarped_to_array(
    #                     self.optimizer_instance.space,
    #                     obs_x[i]
    #             )
    #     obs_x = self.optimizer_instance.transform_sparseArray_to_optSpace(obs_x)
    #     self.optimizer_instance.observe(obs_x, obs_y)
    #     print('成功加载先前观测！')


    def evaluate(self, param):
        return self.function_instance.evaluate(param)

    def run(self):
        pbar = tqdm.tqdm(range(self.n_calls))
        pbar.set_description(f"Optimizer {self.cfg.OPTM.name} is running:")
        
        for ii in pbar:

            tt = time()
            next_points, features = self.optimizer_instance.suggest(self.n_suggestions)  # TODO 1

            # try:
            #     next_points, features = self.optimizer_instance.suggest(self.n_suggestions)  # TODO 1
            # except Exception as e:
            #     # logger.warning("Failure in optimizer suggest. Falling back to random search.")
            #     # logger.exception(e, exc_info=True)
            #     print(json.dumps({"optimizer_suggest_exception": {'iter': ii}}))
            #     # api_config = self.function_instance.get_api_config()
            #     # TODO 直接随机采样
            #     x_guess_configs = self.optimizer_instance.space.sample_configuration(size=self.n_suggestions) # a list
            #     next_points = [x_guess_config.get_dict_unwarped() for x_guess_config in x_guess_configs]
            #     features = [x_guess_config.get_array(sparse=False) for x_guess_config in x_guess_configs]

            suggest_time = time() - tt

            assert len(next_points) == self.n_suggestions, "invalid number of suggestions provided by the optimizer"
            # eval_time = [None for _ in range(self.n_suggestions)]
            function_evals = []
            losses = []
            tt = time()
            for next_point in (next_points):
                # try:
                f_current_eval, loss =  self.evaluate(next_point) # TODO 2
                # except Exception as e:
                #     f_current_eval = np.full((len(self.cfg.TEST_PROBLEM.func_evals),), np.inf, dtype=float)
                #     loss = np.full((len(self.cfg.TEST_PROBLEM.losses),), np.inf, dtype=float)



                function_evals.append(f_current_eval)
                losses.append(loss)
            eval_time = time() - tt
            eval_list = np.asarray(losses)[:, :self.cfg.OPTM.n_obj].ravel().tolist() # TODO
            assert self.cfg.OPTM.n_obj == 1
            tt = time()
            self.optimizer_instance.observe(features, eval_list)  # TODO 3
            # try:
            #     self.optimizer_instance.observe(features, eval_list)  # TODO 3
            # except Exception as e:
                # logger.warning("Failure in optimizer observe. Ignoring these observations.")
                # logger.exception(e, exc_info=True)
                # print(json.dumps({"optimizer_observe_exception": {'iter': ii}}))
            observe_time = time() - tt
            timing = {
                         'suggest_time_per_suggest': suggest_time,
                         'observe_time_per_suggest': observe_time,
                         'eval_time_per_suggest': eval_time
            }
            self.record.append(features, losses, function_evals, timing=timing, suggest_point=next_points)


